#include <cstdint>
#include <gint/clock.h>
#include <gint/display.h>
#include <gint/dma.h>
#include <gint/image.h>
#include <gint/keyboard.h>
#include <gint/rtc.h>

#include <fxlibc/printf.h>

#include "config.h"

#include <gint/kmalloc.h>
#include <libprof.h>

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "utilities/extrakeyboard.h"
#include "utilities/utilities.h"

#include <string>

#if (MORE_RAM)
kmalloc_arena_t extended_ram = {NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                                0,    0,    0,    0,    0,    0,    0};
kmalloc_arena_t *_uram;
kmalloc_gint_stats_t *_uram_stats;
kmalloc_gint_stats_t *extram_stats;
#else
kmalloc_arena_t *_uram;
kmalloc_gint_stats_t *_uram_stats;
#endif

bool exitToOS = false;
bool screenshot = false;
bool record = false;
bool showFPS = true;
bool showDebug = true;

bool canWeAllocateMoreRam = false;

KeyboardExtra Keyboard;

float elapsedTime = 0.0f;
uint32_t time_update = 0, time_render = 0;
prof_t perf_update, perf_render;

extern font_t font_matrix;
extern font_t font_label;
bopti_image_t *screen;

void Initialise(void);
void Update(void);
void Render(void);
void Close(void);

bool AddMoreRAM(void) {
#if (MORE_RAM && !CALCEMU)

  /* allow more RAM */
  char const *osv = (char *)0x80020020;

  if ((!strncmp(osv, "03.", 3) && osv[3] <= '8') &&
      gint[HWCALC] == HWCALC_FXCG50) // CG-50
  {
    extended_ram.name = "extram";
    extended_ram.is_default = true;
    extended_ram.start = (void *)0x8c200000;
    extended_ram.end = (void *)0x8c4e0000;

    kmalloc_init_arena(&extended_ram, true);
    kmalloc_add_arena(&extended_ram);
    return true;
  } else if (gint[HWCALC] == HWCALC_PRIZM) // CG-10/20
  {

    extended_ram.name = "extram";
    extended_ram.is_default = true;

    uint16_t *vram1, *vram2;
    dgetvram(&vram1, &vram2);
    dsetvram(vram1, vram1);

    extended_ram.start = vram2;
    extended_ram.end = (char *)vram2 + 396 * 224 * 2;

    kmalloc_init_arena(&extended_ram, true);
    kmalloc_add_arena(&extended_ram);
    return false;

  } else if (gint[HWCALC] == HWCALC_FXCG_MANAGER) // CG-50 EMULATOR
  {

    extended_ram.name = "extram";
    extended_ram.is_default = true;
    extended_ram.start = (void *)0x88200000;
    extended_ram.end = (void *)0x884e0000;

    kmalloc_init_arena(&extended_ram, true);
    kmalloc_add_arena(&extended_ram);
    return true;
  }
#elif (MORE_RAM && CALCEMU)

  extended_ram.name = "extram";
  extended_ram.is_default = true;
  extended_ram.start = (void *)0x8c200000;
  extended_ram.end = (void *)0x8c4e0000;

  kmalloc_init_arena(&extended_ram, true);
  kmalloc_add_arena(&extended_ram);
  return true;

#else

  return false;

#endif

  return false;
}

void FreeMoreRAM(void) {
#if (MORE_RAM)
  memset(extended_ram.start, 0,
         (char *)extended_ram.end - (char *)extended_ram.start);
#endif
}

void GetInputs([[maybe_unused]] float dt) {
  if (Keyboard.IsKeyPressed(MYKEY_EXIT)) {
    exitToOS = true;
  };

  if (Keyboard.IsKeyPressedEvent(MYKEY_OPTN))
    showFPS = !showFPS;

  if (Keyboard.IsKeyPressedEvent(MYKEY_MENU))
    showDebug = !showDebug;
}

void Initialise(void) {}

void Render(void) {
  if (showFPS) {
    // dfont( &font_label );
    dprint(11, 6, C_RED, "Framerate : %.0f FPS",
           (float)(1000000.0f / elapsedTime));
    dprint(10, 5, C_WHITE, "Framerate : %.0f FPS",
           (float)(1000000.0f / elapsedTime));
  }

  if (showDebug) {
    // dfont( &font_label );

#if (MORE_RAM)
    if (canWeAllocateMoreRam) {
      dprint(261, 6, C_RED, "Mem Used : %d Kb",
             (_uram_stats->used_memory + extram_stats->used_memory) / 1024);
      dprint(260, 5, C_WHITE, "Mem Used : %d Kb",
             (_uram_stats->used_memory + extram_stats->used_memory) / 1024);
      dprint(261, 21, C_RED, "Mem Free : %d Kb",
             (_uram_stats->free_memory + extram_stats->free_memory) / 1024);
      dprint(260, 22, C_WHITE, "Mem Free : %d Kb",
             (_uram_stats->free_memory + extram_stats->free_memory) / 1024);
      dprint(261, 36, C_RED, "Mem Peak : %d Kb",
             (_uram_stats->peak_used_memory + extram_stats->peak_used_memory) /
                 1024);
      dprint(260, 35, C_WHITE, "Mem Peak : %d Kb",
             (_uram_stats->peak_used_memory + extram_stats->peak_used_memory) /
                 1024);
    } else {
      dprint(261, 6, C_RED, "Mem Used : %d Kb",
             (_uram_stats->used_memory) / 1024);
      dprint(260, 5, C_WHITE, "Mem Used : %d Kb",
             (_uram_stats->used_memory) / 1024);
      dprint(261, 21, C_RED, "Mem Free : %d Kb",
             (_uram_stats->free_memory) / 1024);
      dprint(260, 20, C_WHITE, "Mem Free : %d Kb",
             (_uram_stats->free_memory) / 1024);
      dprint(261, 36, C_RED, "Mem Peak : %d Kb",
             (_uram_stats->peak_used_memory) / 1024);
      dprint(260, 35, C_WHITE, "Mem Peak : %d Kb",
             (_uram_stats->peak_used_memory) / 1024);
    }
#else
    dprint(261, 6, C_RED, "Mem Used : %d Kb",
           (_uram_stats->used_memory) / 1024);
    dprint(260, 5, C_WHITE, "Mem Used : %d Kb",
           (_uram_stats->used_memory) / 1024);
    dprint(261, 21, C_RED, "Mem Free : %d Kb",
           (_uram_stats->free_memory) / 1024);
    dprint(260, 20, C_WHITE, "Mem Free : %d Kb",
           (_uram_stats->free_memory) / 1024);
    dprint(261, 36, C_RED, "Mem Peak : %d Kb",
           (_uram_stats->peak_used_memory) / 1024);
    dprint(260, 35, C_WHITE, "Mem Peak : %d Kb",
           (_uram_stats->peak_used_memory) / 1024);
#endif
  }
}

void Update([[maybe_unused]] float dt) {}

void Close(void) {}

int main(void) {

  _uram = kmalloc_get_arena("_uram");
#if (MORE_RAM)
  canWeAllocateMoreRam = AddMoreRAM();
#endif

  int EntryClockLevel;
  EntryClockLevel = clock_get_speed();
  clock_set_speed(CLOCK_SPEED_F4);

  srand(rtc_ticks());

  exitToOS = false;

  Initialise();

  __printf_enable_fp();
  __printf_enable_fixed();

  prof_init();

  do {
    perf_update = prof_make();
    prof_enter(perf_update);
    {
      // all the stuff to be update should be put here
      Keyboard.Update();

      // read inputs from the player
      GetInputs(elapsedTime);

      // update as per the time spend to do the loop
      Update(elapsedTime);

      // update the RAM consumption status
      _uram_stats = kmalloc_get_gint_stats(_uram);
#if (MORE_RAM)
      extram_stats = kmalloc_get_gint_stats(&extended_ram);
#endif
    }
    prof_leave(perf_update);
    time_update = prof_time(perf_update);

    perf_render = prof_make();
    prof_enter(perf_render);
    {
      dclear(C_BLACK);

      // Call the render functions
      Render();

      dupdate();
    }
    prof_leave(perf_render);
    time_render = prof_time(perf_render);

    elapsedTime = ((float)(time_update + time_render));

  } while (exitToOS == false);

  Close();

  prof_quit();

  clock_set_speed(EntryClockLevel);

#if (MORE_RAM)
  if (canWeAllocateMoreRam)
    FreeMoreRAM();
#endif

  return 1;
}
